#include "NNEstimator.h"
#include "utils.h"
#include "globals.h"
#include "io_utils.h"
#include <time.h>
#include <stdlib.h>
#include <iostream>
#include <string>
#include <sstream>
#include <ctime>
#include <cmath>
#include <algorithm>


using namespace std;

NNEstimator::NNEstimator(const char* dir, int db_pages, int dist_type) {
	this->dir = dir;
	this->db_pages = db_pages;
	this->ts_size = 0;
	this->page_size = -1;
	this->dist_type = dist_type;
	loadInds();
	loadExamples();
	string strdir = dir;
    LOG.open((strdir + "Output_indsNND_out.txt").c_str());
}

NNEstimator::~NNEstimator() {
	cleanContainer(inds_dists_nn);
	cleanContainer(sampleDb);
	LOG.close();
}

void NNEstimator::computeNNDists() {
	vector <float*> DB;
	int starttimer = clock();
	int offset = 0;
	for (int i = 0; i < db_pages; i++){
        cout << "---(NNDists for sample) Checking page#" << (i+1) << " of " << db_pages << endl;
        loadDB(DB, i);
		updateNNbs(DB, offset);
        offset += (int) DB.size();
		cleanContainer(DB);
		if (db_pages > 1) {
			this->page_size = -1;
			ts_sizeconstant = this->ts_size;
			this->ts_size = -1;
		}
	}
	int endtimer = clock();
	double time_spent = (double) (endtimer-starttimer) / (double) 1000;
	LOG << "Time to compute NNDists = " << time_spent << "sec." << endl;
	logNbs();
}

void NNEstimator::logNbs() {
	for (int i = 0; i < sampleSize; i++) {
		float dist = (float) sqrt(inds_dists_nn[i][1]);
		LOG << inds_dists_nn[i][0] << "\t" << dist << "\t" << inds_dists_nn[i][2] << endl;
	}
}

void NNEstimator::updateNNbs(vector<float*> &DB, int offset) {
	for (int i = 0; i < (int) DB.size(); i++) {
		float* dbEl = DB[i];
		//cout << "Ts #" << i << endl;
		for (int j = 0; j < sampleSize; j++) {
			float ind = inds_dists_nn[j][0];
			float dst = inds_dists_nn[j][1];
			if (offset + i == ind) { //the main memory element is the current DB element that we check. Skip it.
				continue;
			}
			double dist = nnDist(dbEl, sampleDb[j], true, dst);
			if (dist < dst) {
				inds_dists_nn[j][1] = (float) dist;
				inds_dists_nn[j][2] = (float) (offset+i);
			}
		}
	}
}

double NNEstimator::nnDist(float* x, float* y, bool earlyAbandon, double minDist) {
	double* dist_numoper = NULL;
	switch (dist_type) {
		case Globals::EUCLID_DIST : {
			dist_numoper = sqreuclid_distf(x, y, ts_size, earlyAbandon, minDist);
			break;
		}
		case Globals::ROTAT_DIST: {
			dist_numoper = sqr_rotatdist(x, y, ts_size, earlyAbandon, minDist);
			break;
		}
	}
	//todo: store the number of performed operations
	double dist = dist_numoper[0];
	delete [] dist_numoper;
	return dist;
}

void NNEstimator::loadDB(vector<float*>& DB, int pageNum) {
	// generate the file name
	string fileName(dir);
	fileName += "TS";
	stringstream oss;
    oss << fileName << pageNum;
	fileName = oss.str();
	fileName += ".txt";

	read_matrix1(DB, fileName, page_size, ts_size);
}

void NNEstimator::loadInds() {
	string fileName = dir;
    fileName += "Output_indsNND.txt";
	
    char str[200];
    fstream file(fileName.c_str(), ios::in);
	sampleSize = 0;
	vector<int> inds;
    while(!file.eof())  {
		file.getline(str,200);
		int ind = atoi(str);
		inds.push_back(ind);
		sampleSize++;
	}
	file.close();
	//sort inds
	sort(inds.begin(), inds.end());
	for (int i = 0; i < sampleSize; i++) {
		float* tmp_ind_dst_nn = new float[3];
		tmp_ind_dst_nn[0] = (float) inds[i];
		tmp_ind_dst_nn[1] = 1000000000;
		tmp_ind_dst_nn[2] = -1;
		inds_dists_nn.push_back(tmp_ind_dst_nn);
	}
}

void NNEstimator::loadExamples() {
	int starttimer = clock();
	cout << "In build sample...." << endl;
    int offset = 0;
	int position = 0;
	int linecounter = 0;
	for (int i = 0; i < db_pages; i++){
        cout << "---Extracting samples from page#" << (i+1) << " of " << db_pages << endl;
        loadSample(i, position, linecounter);
 	}
	cout << "sample DB.size=" << (int) sampleDb.size() << endl;
	int endtimer = clock();
	double time_spent = (double) (endtimer-starttimer) / (double) 1000;
	cout << "Time to build sample: " << time_spent << endl;
}

void NNEstimator::loadSample(int pageNum, int& position, int& lineCounter) {
	// generate the file name
	string fileName(dir);
	fileName += "TS";
	stringstream oss;
    oss << fileName << pageNum;
	fileName = oss.str();
	fileName += ".txt";

	FILE* fp = fopen(fileName.c_str(), "r");
	long size = file_size(fp);
	fseek(fp, 0, SEEK_SET);
	char* buf = new char[size];
	fread((void*) buf, sizeof(char), size, fp);
	fclose(fp);

	char *st, sep[]="\n";
    st=strtok(buf,sep);
	int i = 0;
	while (st != NULL) {
		if (ts_size == 0) {
			setTsSize(st);
		}
		if ((int) strlen(st) > ts_size && position < this->sampleSize) {
			if (lineCounter == inds_dists_nn[position][0]) {
				float* example = readExample(st);
				sampleDb.push_back(example);
				position++;
			}
			lineCounter++;
		}
		st=strtok(NULL,sep);
	}
	delete [] buf;
}

void NNEstimator::setTsSize(char* cline) {
	int pos1 = 0; 
	int pos2 = 0;
	while (cline[pos1] && (cline[pos1] == ' ' || cline[pos1] == '\t')) {
		pos1++;
	}
	pos2 = pos1;
	while (cline[pos2]) {
		while (cline[pos2] && cline[pos2] != ' ' && cline[pos2] != '\t') {
			pos2++;
		}
		ts_size++;
		pos1 = pos2;
		while (cline[pos2] && (cline[pos2] == ' ' || cline[pos2] == '\t')) {
			pos2++;
		}
	}
}

float* NNEstimator::readExample(char* cline) {
	int pos1 = 0; 
	int pos2 = 0;
	float* ts = new float[ts_size];
	int j = 0;
	while (cline[pos1] && (cline[pos1] == ' ' || cline[pos1] == '\t')) {
		pos1++;
	}
	pos2 = pos1;
	while (cline[pos2]) {
		while (cline[pos2] && cline[pos2] != ' ' && cline[pos2] != '\t') {
			pos2++;
		}
		char* num = new char[pos2-pos1 + 1];
		for (int i = 0; i < (pos2-pos1); i++) {
			num[i] = cline[pos1+i];
		}
		num[pos2-pos1] = 0;
		float l= (float) atof(num);
		ts[j++] = l;
		delete [] num;
		pos1 = pos2;
		while (cline[pos2] && (cline[pos2] == ' ' || cline[pos2] == '\t')) {
			pos2++;
		}
	}
	return ts;
}