#include "test.h"
#include "io_utils.h"
#include <stdlib.h>

using namespace std;

void Test::run() {
	// check io speed for external and internal drive
	//runIOTest();
	// check how adding and deleting from a container performs
	runContainerTest();
}

void Test::runContainerTest() {
	int starttimer = clock();
	//add a million items to a vector and then erase half of them
	vector<float*> container;
	cout << "Loading the container." << endl;
	for (int i = 0; i < 1000000; i++) {
		float* item = new float[1];
		container.push_back(item);
	}
	cout << "Unloading the container." << endl;
	for (int i = 0; i < 500000; i++) {
	//draw a uniformly random sample
		srand( (unsigned)time( NULL ) );
		int rnd1 = rand();
		int rnd2 = rand();  //necessary because the maximal random number is 32767
		int rnd = rnd1*rnd2;
		int inds = rnd % (int) container.size();
		float* el = container[inds];
		container.erase(container.begin() + inds);
		delete [] el;
	}
	cleanFloatContainer(container);
	int endtimer = clock();
	double time_spent = (double) (endtimer-starttimer) / (double) 1000;
	cout << "(Phase2) Time to find discords: " << time_spent << "sec." << endl;
}

void Test::runIOTest() {
	//checking reading time from external drive
	dir = "E:\\data\\randomwalk_1mill_size512\\";
	checkReadingTime();
	//checking reading time from internal drive
	dir = "C:\\Projects\\discdiscords\\data\\randomwalk\\";
	checkReadingTime();
}

void Test::checkReadingTime() {
	int starttimer = clock();
	vector <float*> DB;
	int db_pages = 1;
	for (int i = 0; i < db_pages; i++){
        cout << "--- Checking page#" << (i+1) << " of " << db_pages << endl;
        loadDB(DB, i);
		cleanFloatContainer(DB);
	}
	int endtimer = clock();
	double time_spent = (double) (endtimer-starttimer) / (double) 1000;
	cout << "(" << dir << ")Reading time: " << time_spent << "sec." << endl;
}

void Test::loadDB(vector<float*>& DB, int pageNum) {
	// generate the file name
	string fileName(dir);
	fileName += "TS";
	stringstream oss;
    oss << fileName << pageNum;
	fileName = oss.str();
	fileName += ".txt";
	int page_size = -1;
	int ts_len = -1;
	read_matrix(DB, fileName, page_size, ts_len);
}

void Test::cleanFloatContainer(vector <float*>& cont) {
	while (cont.size() > 0) {
		float* vEl = cont[0];
		cont.erase(cont.begin());
		delete [] vEl;
	}
}