#include "threshold_estimator.h"
#include "utils.h"
#include "io_utils.h"
#include "disc_discords.h"
#include "globals.h"
#include <time.h>
#include <stdlib.h>
#include <iostream>
#include <string>
#include <sstream>
#include <ctime>
#include <cmath>
#include <algorithm>

using namespace std;

ThresholdEstimator::ThresholdEstimator(const char* dir, const char* file, double db_size, 
									   double sampleSize, int db_pages, int dist_type, double num_discords) {
	this->db_size = db_size;
	this->threshold = 0;
	this->dir = dir;
	this->file = file;
	this->sampleSize = sampleSize;
	this->db_pages = db_pages;
	this->position = 0;
	this->lineCounter = 0;
	this->ts_size = 0;
	this->page_size = -1;
	this->dist_type = dist_type;
	this->num_discords = num_discords;
	this->usePivots = Globals::USE_PIVOTS;
	string strdir = dir;
    LOG.open((strdir + "Output_thresholdestimator.txt").c_str());
	if (usePivots) {
		PIVOTLOG.open((strdir + "Output_pivots.txt").c_str());
	}
}

ThresholdEstimator::~ThresholdEstimator() {
	cleanContainer(sample);
	cleanContainer(sampleDb);
	cleanContainer(pivots);
	cleanContainer(nnSampleDists);
    LOG.close();
	if (usePivots) {
		PIVOTLOG.close();
	}
}

// This computes the threshold based only on samples, without scanning the entire database
void ThresholdEstimator::computeThreshold1() {
	string strdir = dir;
    LOG1.open((strdir + "Output_thresholdestimator1.txt").c_str());
	//draw a uniformly random sample
	int starttimer = clock();
	int numsamples = 1;
	for (int i = 0; i < numsamples; i++) {
		computeRndInds();
		buildSample();
		computeSampleNNDists();
		for (int j = 0; j < sampleSize; j++) {
			float minDist = 100000000;
			for (int p = 0; p < sampleSize; p++) {
				if (j != p && nnSampleDists[j][p] < minDist) {
					minDist = nnSampleDists[j][p];
				}
			}
			sample[j][1] = minDist;
		}
		sortSamplesDists();
		for (int j = (int) sampleSize-1; j >= 0; j--) {
			LOG1 << "" << sample[j][0] << " " << sample[j][1] << endl; 	
		}
		logPivots();
		this->threshold += sample[(int) sampleSize-1][1];
		cleanContainer(sample);
		cleanContainer(sampleDb);
		cleanContainer(nnSampleDists);
		this->position = 0;
		this->lineCounter = 0;
	}
	this->threshold /= (float) numsamples;
	LOG1.close();
	int endtimer = clock();
	double time_spent = (double) (endtimer-starttimer) / (double) 1000;
	LOG << "Time to compute threshold = " << time_spent << "sec." << endl;
	LOG << "THRESHOLD = " << this->threshold << endl;
}

void ThresholdEstimator::computeThreshold() {
	string strdir = dir;
    LOG1.open((strdir + "Output_thresholdestimator2_nnd.txt").c_str());
	//draw a uniformly random sample
	int starttimer = clock();
	computeRndInds();
	buildSample();
	this->ts_size = -1;
	int endtimer = clock();
	double time_spent = (double) (endtimer-starttimer) / (double) 1000;
	LOG << "Time to extract sample elements = " << time_spent << "sec." << endl;
	starttimer = clock();
	//find the nearest neighbor distance to each element in the sample
	computeNNDists();
	for (int j = (int) sampleSize-1; j >= 0; j--) {
		LOG1 << "" << sample[j][0] << " " << sample[j][1] << endl; 	
	}
	LOG1.close();
	//sort the distances and find the distance threshold
	computeThresholdFromDists();
	
	ts_size = ts_sizeconstant;
	logPivots();
	endtimer = clock();
	time_spent = (double) (endtimer-starttimer) / (double) 1000;
	LOG << "Time to compute NN dists to elements in sample = " << time_spent << "sec." << endl;
	LOG << "Data base size = " << db_size << endl;
	LOG << "Sample size = " << sampleSize << endl;
	LOG << "DB pages = " << db_pages << endl;
	LOG << "Number of discords to be detected in DB = " << num_discords << endl;
}

void ThresholdEstimator::logPivots() {
	if (!usePivots) {
		return;
	}
	//collectGoodPivots();
	collectGoodPivots1(100);
	for (int i = 0; i < (int) pivots.size(); i++) {
		for (int j = 0; j < this->ts_size; j++) {
			PIVOTLOG << "" << pivots[i][j] << " ";
		}
		PIVOTLOG << endl;
	}
}

//collects pivots as uniformly random sample
void ThresholdEstimator::collectGoodPivots1(int numPivots) {
	vector<double> dbinds;
	for (int i = 0; i < sampleSize; i++) {
		dbinds.push_back(i);
	}
	vector<float*> smplInds;
	//draw a uniformly random sample
	srand( (unsigned)time( NULL ) );
	for (int i = 0; i < numPivots; i++) {
		int rnd1 = rand();
		int rnd2 = rand();  //necessary because the maximal random number is 32767
		int rnd = rnd1*rnd2;
		int inds = rnd % (int) dbinds.size();
		addToSampleInds((int) dbinds[inds], smplInds);
		dbinds.erase(dbinds.begin() + inds);
	}
	dbinds.clear();
	for (int i = 0; i < numPivots; i++) {
		int inds = (int) smplInds[i][0];
		float* pivot = new float[this->ts_size];
		for (int j = 0; j < ts_size; j++) {
			pivot[j] = sampleDb[inds][j];
		}
		pivots.push_back(pivot);
	}
}

// collects as pivots points that are on average close to all others
// and very far from all others too
void ThresholdEstimator::collectGoodPivots() {
	// Pivot1: the time series with smallest NN dist
	// Pivot2: the time series with largest NN dist (do we need it?)
	float* pivot1 = new float[this->ts_sizeconstant];
	float* pivot2 = new float[this->ts_sizeconstant];
	for (int i = 0; i < this->ts_sizeconstant; i++) {
		pivot1[i] = sampleDb[0][i];
		pivot2[i] = sampleDb[(int)sampleSize-1][i];
	}
	cout << "pivot1 index: " << sample[0][0] << endl;
	cout << "pivot2 index: " << sample[(int)sampleSize-1][0] << endl;
	pivots.push_back(pivot1);
	pivots.push_back(pivot2);
	// Pivot3: the time series with smallest average NN sample dist
	computeSampleNNDists();
	float* pivot3 = minAvrDistPivot();
	pivots.push_back(pivot3);
	// Pivot4: the time series furthest from Pivot1 in the sample
	float* pivot4 = furthestFromPivot(0);
	pivots.push_back(pivot4);
}

float* ThresholdEstimator::furthestFromPivot(int pivotInd) {
	float* dists = nnSampleDists[pivotInd];
	int indx = 0;
	float maxDist = 0;
	for (int i = 0; i < sampleSize; i++) {
		if (dists[i] > maxDist) {
			indx = i;
			maxDist = dists[i];
		}
	}
	cout << "Pivot3 index: " << sample[indx][0] << endl;
	float* pivot = new float[ts_sizeconstant];
	for (int i = 0; i < ts_sizeconstant; i++) {
		pivot[i] = sampleDb[indx][i];
	}
	return pivot;
}

float* ThresholdEstimator::minAvrDistPivot() {
	float* avrDists = new float[(int) sampleSize];
	for (int i = 0; i < sampleSize; i++) {
		float tmpDist = 0;
		for (int j = 0; j < sampleSize; j++) {
			tmpDist += nnSampleDists[i][j];
		}
		avrDists[i] = (tmpDist / (float) sampleSize);
	}
	int minInds = 0;
	float minAvrDist = avrDists[0];
	for (int i = 1; i < sampleSize; i++) {
		if (minAvrDist > avrDists[i]) {
			minAvrDist = avrDists[i];
			minInds = i;
		}
	}
	cout << "Pivot4 index: " << sample[minInds][0] << endl;
	float* pivot = new float[ts_sizeconstant];
	for (int i = 0; i < ts_sizeconstant; i++) {
		pivot[i] = sampleDb[minInds][i];
	}
	return pivot;
}

void ThresholdEstimator::computeSampleNNDists() {
	for (int i = 0; i < sampleSize; i++) {
		float* dists = new float[(int) sampleSize];
		for (int j = 0; j < sampleSize; j++) {
			if (i > j) {
				dists[j] = nnSampleDists[j][i];
			} else if (i == j) {
				dists[j] = 0;
			} else {
				dists[j] = (float) sqrt(nnDist(sampleDb[i], sampleDb[j], false, 100000000));
			}
		}
		nnSampleDists.push_back(dists);
	}
}

void ThresholdEstimator::computeThresholdFromDists() {
	//todo: may be true for other distances too
	if (this->dist_type == Globals::EUCLID_DIST || this->dist_type == Globals::ROTAT_DIST) {
		for (int i = 0; i < sampleSize; i++) {
			sample[i][1] = sqrt(sample[i][1]);
		}
	}
	//order the NN distances d1<=d2<=...<=dm
	sortSamplesDists();
	for (int i = 0; i < sampleSize; i++) {
		cout << "dist = " << sample[i][1] << endl;
	}
	//use Chernoff bound to compute
	computeEpsilon();
	double tail_volume = num_discords/db_size;
	double num_discords_sample = round(tail_volume*sampleSize);
	if (num_discords_sample < 1) {
		num_discords_sample = 1;
	}
	double sample_tail_count = num_discords_sample + this->epsilon;
	double last_nontail = sampleSize - sample_tail_count;
	this->threshold = (float) sample[(int) last_nontail][1];
	LOG << "Computing the Chernoff bound using the description in Scholkopf&Smola(Lerning with Kernels)" << endl;
	LOG << "Epsilon = " << this->epsilon << endl;
	LOG << "Elements in tail = " << sample_tail_count << "(out of " << sampleSize << ")" << endl;
	LOG << "THRESHOLD = " << this->threshold << endl;
}

void ThresholdEstimator::sortSamplesDists() {
	for (int i = 0; i < sampleSize; i++) {
		int minInds = 0;
		float minDist = 100000000;
		for (int j = i; j < sampleSize; j++) {
			if (sample[j][1] < minDist) {
				minDist = sample[j][1];
				minInds = j;
			}
		}
		if (i != minInds) {
			float* tmpIndsDist = sample[minInds];
			float* tmpSample = sampleDb[minInds];
			sample[minInds] = sample[i];
			sample[i] = tmpIndsDist;
			sampleDb[minInds] = sampleDb[i];
			sampleDb[i] = tmpSample;
		}
	}
}

// Use Chernoff bound to compute epsilon to ensure that with
// .99 percent probability the threshold will cut the required volume
// from the true distribution
// if m == sampleSize
// 2exp(-2meps^2) <= 0.01
void ThresholdEstimator::computeEpsilon() {
	double perc_epsilon = sqrt(log((float) 200) / (float) (2*sampleSize));
	this->epsilon = round(perc_epsilon*sampleSize);
	if (this->epsilon < 1) {
		this->epsilon = 1;
	}
}

void ThresholdEstimator::computeNNDists() {
	vector <float*> DB;
	int offset = 0;
	for (int i = 0; i < db_pages; i++){
        cout << "---(NNDists for sample) Checking page#" << (i+1) << " of " << db_pages << endl;
        loadDB(DB, i);
		updateNNbs(DB, offset);
        offset += (int) DB.size();
		cleanContainer(DB);
		if (db_pages > 1) {
			this->page_size = -1;
			ts_sizeconstant = this->ts_size;
			this->ts_size = -1;
		}
	}
}

void ThresholdEstimator::updateNNbs(vector<float*> &DB, int offset) {
	for (int i = 0; i < (int) DB.size(); i++) {
		float* dbEl = DB[i];
		vector <float*>::iterator s_Iter;
		vector <float*>::iterator indxPtrIter = sample.begin();
		//cout << "Ts #" << i << endl;
		for (s_Iter = sampleDb.begin(); s_Iter != sampleDb.end(); s_Iter++) {
			float inds = (*indxPtrIter)[0];
			float dst = (*indxPtrIter)[1];
			if (offset + i == (*indxPtrIter)[0]) { //the main memory element is the current DB element that we check. Skip it.
				indxPtrIter++;
				continue;
			}
			double dist = nnDist(dbEl, *s_Iter, true, (*indxPtrIter)[1]);
			if (dist < (*indxPtrIter)[1]) {
				(*indxPtrIter)[1] = (float) dist;
			}
			indxPtrIter++;
		}
	}
}

void ThresholdEstimator::buildSample() {
	int starttimer = clock();
	cout << "In build sample...." << endl;
    int offset = 0;
	for (int i = 0; i < db_pages; i++){
        cout << "---Extracting samples from page#" << (i+1) << " of " << db_pages << endl;
        loadSample(i);
 	}
	cout << "sample DB.size=" << (int) sampleDb.size() << endl;
	int endtimer = clock();
	double time_spent = (double) (endtimer-starttimer) / (double) 1000;
	cout << "Time to build sample: " << time_spent << endl;
}

void ThresholdEstimator::loadSample(int pageNum) {
	// generate the file name
	string fileName(dir);
	fileName += "TS";
	stringstream oss;
    oss << fileName << pageNum;
	fileName = oss.str();
	fileName += ".txt";

	FILE* fp = fopen(fileName.c_str(), "r");
	long size = file_size(fp);
	fseek(fp, 0, SEEK_SET);
	char* buf = new char[size];
	fread((void*) buf, sizeof(char), size, fp);
	fclose(fp);

	char *st, sep[]="\n";
    st=strtok(buf,sep);
	int i = 0;
	while (st != NULL) {
		if (ts_size == 0) {
			setTsSize(st);
		}
		if ((int) strlen(st) > ts_size && position < this->sampleSize) {
			if (lineCounter == sample[position][0]) {
				float* example = readExample(st);
				sampleDb.push_back(example);
				position++;
			}
			lineCounter++;
		}
		st=strtok(NULL,sep);
	}
	delete [] buf;
}

void ThresholdEstimator::setTsSize(char* cline) {
	int pos1 = 0; 
	int pos2 = 0;
	while (cline[pos1] && (cline[pos1] == ' ' || cline[pos1] == '\t')) {
		pos1++;
	}
	pos2 = pos1;
	while (cline[pos2]) {
		while (cline[pos2] && cline[pos2] != ' ' && cline[pos2] != '\t') {
			pos2++;
		}
		ts_size++;
		pos1 = pos2;
		while (cline[pos2] && (cline[pos2] == ' ' || cline[pos2] == '\t')) {
			pos2++;
		}
	}
}

float* ThresholdEstimator::readExample(char* cline) {
	int pos1 = 0; 
	int pos2 = 0;
	float* ts = new float[ts_size];
	int j = 0;
	while (cline[pos1] && (cline[pos1] == ' ' || cline[pos1] == '\t')) {
		pos1++;
	}
	pos2 = pos1;
	while (cline[pos2]) {
		while (cline[pos2] && cline[pos2] != ' ' && cline[pos2] != '\t') {
			pos2++;
		}
		char* num = new char[pos2-pos1 + 1];
		for (int i = 0; i < (pos2-pos1); i++) {
			num[i] = cline[pos1+i];
		}
		num[pos2-pos1] = 0;
		float l= (float) atof(num);
		ts[j++] = l;
		delete [] num;
		pos1 = pos2;
		while (cline[pos2] && (cline[pos2] == ' ' || cline[pos2] == '\t')) {
			pos2++;
		}
	}
	return ts;
}

void ThresholdEstimator::computeRndInds() {
	vector<double> dbinds;
	for (int i = 0; i < db_size; i++) {
		dbinds.push_back(i);
	}
	//draw a uniformly random sample
	srand( (unsigned)time( NULL ) );
	for (int i = 0; i < sampleSize; i++) {
		int rnd1 = rand();
		int rnd2 = rand();  //necessary because the maximal random number is 32767
		int rnd = rnd1*rnd2;
		int inds = rnd % (int) dbinds.size();
		addToSampleInds((int) dbinds[inds], sample);
		dbinds.erase(dbinds.begin() + inds);
	}
	dbinds.clear();
}

void ThresholdEstimator::addToSampleInds(int inds, vector<float*>& smpl) {
	vector<float*>::iterator iter = smpl.begin();
	float* ind_nndist = new float[2];
	ind_nndist[0] = (float) inds;
	ind_nndist[1] = (float) 10000000;
	while (iter != smpl.end() && (*iter)[0] < inds) {
		iter++;
	}
	if (iter == smpl.end()) {
		smpl.push_back(ind_nndist);
	} else {
		smpl.insert(iter, ind_nndist);
	}
}

void ThresholdEstimator::loadDB(vector<float*>& DB, int pageNum) {
	// generate the file name
	string fileName(dir);
	fileName += "TS";
	stringstream oss;
    oss << fileName << pageNum;
	fileName = oss.str();
	fileName += ".txt";

	read_matrix1(DB, fileName, page_size, ts_size);
}

double ThresholdEstimator::nnDist(float* x, float* y, bool earlyAbandon, double minDist) {
	double* dist_numoper = NULL;
	switch (dist_type) {
		case Globals::EUCLID_DIST : {
			dist_numoper = sqreuclid_distf(x, y, ts_size, earlyAbandon, minDist);
			break;
		}
		case Globals::ROTAT_DIST: {
			dist_numoper = sqr_rotatdist(x, y, ts_size, earlyAbandon, minDist);
			break;
		}
	}
	//todo: store the number of performed operations
	double dist = dist_numoper[0];
	delete [] dist_numoper;
	return dist;
}
